#ifndef IDEEP_COMPUTATIONS_HPP
#define IDEEP_COMPUTATIONS_HPP

#include "lru_cache.hpp"
#include "operators/batchnorm.hpp"
#include "operators/binary.hpp"
#include "operators/channel_shuffle.hpp"
#include "operators/concat.hpp"
#include "operators/conv.hpp"
#include "operators/deconv.hpp"
#include "operators/direct_copy.hpp"
#include "operators/dropout.hpp"
#include "operators/eltwise.hpp"
#include "operators/gru.hpp"
#include "operators/inner_product.hpp"
#include "operators/layernorm.hpp"
#include "operators/lbr_gru.hpp"
#include "operators/lrn.hpp"
#include "operators/lstm.hpp"
#include "operators/matmul.hpp"
#include "operators/pool.hpp"
#include "operators/prelu.hpp"
#include "operators/softmax.hpp"
#include "operators/spliter.hpp"
#include "operators/sum.hpp"
#include "operators/vanilla_rnn.hpp"

#endif
